import importlib
from multiprocessing import Process
from typing import List, Optional

from typer import Argument, Option, Typer, echo

from snok import __version__
from snok.const import (
    APP_DESC,
    APP_NAME,
    BASE_DEPS,
    BASE_DEV_DEPS,
    BumpType,
    ContentType,
    DepencencyAction,
    ProjectType,
)
from snok.services.generator import (
    _ModelContentGenerator,
    _RouterContentGenerator,
    _ScaffoldContentGenerator,
)
from snok.services.new import NewAppService, NewPackageService
from snok.utils import (
    _bump_version_string,
    _get_default_output_dir,
    _get_project_name,
    _run_cmd,
    _snok_sources,
    _update_pyproject_toml,
)

app = Typer(
    name=APP_NAME,
    help=APP_DESC,
    no_args_is_help=True,
)


@app.callback()
def main_callback() -> None:
    pass


main_callback.__doc__ = APP_NAME


@app.command("version")
def _version() -> None:
    """Prints the version of snok."""
    echo(__version__, nl=False)


@app.command(
    "new",
    help="Create various types of Python projects.",
    no_args_is_help=True,
)
def _new(
    name: str = Argument(
        ...,
        help="The name of the project.",
    ),
    description: str = Option(
        "A new Python project.",
        "-d",
        "--description",
        help="The description of the project. Defaults to 'A new Python project.'",
    ),
    type: ProjectType = Option(
        "package",
        "-t",
        "--type",
        help="The type of project to create. Defaults to 'package'",
    ),
    output_dir: str = Option(
        _get_default_output_dir(),
        "-o",
        "--output-dir",
        help="The output directory of the project. "
        "Defaults to the same directory of the current virtual environment.",
    ),
) -> None:
    new_project_service_dispatcher = {
        ProjectType.package: NewPackageService,
        ProjectType.app: NewAppService,
    }
    new_project_service = new_project_service_dispatcher[type]()
    echo(f"Creating project {name}...")
    new_project_service.create(
        name=name,
        desc=description,
        type=type,
        output_dir=output_dir,
    )
    if type == ProjectType.app:  # pragma: no cover
        echo("Adding project dependencies...")
        _add(packages=BASE_DEPS, dependency_group=None, install=False)
    echo("Adding project dev dependencies...")
    _add(packages=BASE_DEV_DEPS, dependency_group="dev")


@app.command(
    "clean",
    help="Clean the project. Power-washers not included.",
)
def _clean() -> None:  # pragma: no cover
    _run_cmd(
        [
            "rm",
            "-rf",
            "build",
            "dist",
            "site",
            "htmlcov",
            ".mypy_cache",
            ".pytest_cache",
            ".ruff_cache",
            "*.egg-info",
            "coverage.xml",
            ".coverage",
        ]
    )


@app.command(
    "install",
    help="Install the project.",
)
def _install(
    editable: bool = Option(
        True,
        "-e",
        "--editable",
        help="Install the project in editable mode. Defaults to True.",
    ),
    dependency_groups: Optional[List[str]] = Option(
        ["dev"],
        "-g",
        "--dependency-groups",
        help="The dependency groups to install. Defaults to ['dev'].",
    ),
) -> None:  # pragma: no cover
    _run_cmd(
        ["pip", "install", "--upgrade", "pip"],
    )
    dep_install_str = "."
    if dependency_groups:
        dep_install_str += f"[{','.join(dependency_groups)}]"
    _run_cmd(
        [
            "pip",
            "install",
            "-e" if editable else "",
            dep_install_str,
        ]
    )
    _run_cmd(
        ["pre-commit", "install"],
    )
    _run_cmd(
        ["pre-commit", "autoupdate"],
    )
    _run_cmd(
        [
            "if",
            "command",
            "-v",
            "pyenv",
            "1>/dev/null",
            "2>&1;",
            "then",
            "pyenv",
            "rehash;",
            "fi",
        ],
    )


@app.command(
    "add",
    help="Add packages to the project.",
    no_args_is_help=True,
)
def _add(
    packages: List[str] = Argument(
        ...,
        help="The packages to add.",
    ),
    dependency_group: Optional[str] = Option(
        None,
        "-g",
        "--dependency-group",
        help="The dependency group to add the packages to. Defaults to None.",
    ),
    install: bool = Option(
        True,
        "-i",
        "--install",
        help="Install the packages after adding them. Defaults to True.",
    ),
) -> None:  # pragma: no cover
    _update_pyproject_toml(
        packages=packages,
        dependency_action=DepencencyAction.add,
        dependency_group_name=dependency_group,
    )
    if install:
        _install(dependency_groups=[dependency_group] if dependency_group else ["dev"])


@app.command(
    "remove",
    help="Remove packages from the project.",
    no_args_is_help=True,
)
def _remove(
    packages: List[str] = Argument(
        ...,
        help="The packages to remove.",
    ),
    dependency_group: str = Option(
        None,
        "-g",
        "--dependency-group",
        help="The dependency group to remove the packages from. Defaults to None.",
    ),
    install: bool = Option(
        True,
        "-i",
        "--install",
        help="Install the packages after removing them. Defaults to True.",
    ),
) -> None:  # pragma: no cover
    _update_pyproject_toml(
        packages=packages,
        dependency_action=DepencencyAction.remove,
        dependency_group_name=dependency_group,
    )
    _run_cmd(
        ["pip", "uninstall", "-y", *packages],
    )
    if install:
        _install(dependency_groups=[dependency_group] if dependency_group else ["dev"])


@app.command(
    "lint",
    help="Check typing and formatting.",
)
def _lint() -> None:  # pragma: no cover
    echo("Linting...")
    _run_cmd(
        ["mypy", *_snok_sources()],
    )
    _run_cmd(
        ["black", "--check", *_snok_sources()],
    )
    _run_cmd(
        ["ruff", *_snok_sources()],
    )


@app.command(
    "format",
    help="Format the code.",
)
def _format(
    extra_paths: Optional[List[str]] = Argument(
        None,
        help="Extra paths to format. Defaults to [].",
    ),
) -> None:  # pragma: no cover
    if not isinstance(extra_paths, list):
        extra_paths = []
    echo("Formatting...")
    _run_cmd(
        ["black", *_snok_sources(), *extra_paths],
    )
    _run_cmd(
        ["ruff", "--fix", *_snok_sources(), *extra_paths],
    )


@app.command(
    "test",
    help="Run the tests.",
)
def _test(
    keepitonehundred: bool = Option(
        False,
        "--keepitonehundred",
        help="Fail if the test coverage is less than 100%. Defaults to False.",
    )
) -> None:  # pragma: no cover
    echo("Running tests...")
    cmd = ["pytest"]
    if keepitonehundred:
        cmd.append("--cov-fail-under=100")
    _run_cmd(cmd)


@app.command(
    "build",
    help="Build the project.",
)
def _build() -> None:  # pragma: no cover
    _run_cmd(
        ["pip", "install", "--upgrade", "build"],
    )
    _run_cmd(
        ["python", "-m", "build"],
    )


@app.command(
    "publish",
    help="Publish the project.",
)
def _publish() -> None:  # pragma: no cover
    _run_cmd(
        ["pip", "install", "--upgrade", "twine"],
    )
    _run_cmd(
        ["twine", "upload", "dist/*"],
    )


@app.command(
    "ok",
    help="Run a suite of commands to make things ok for local dev.",
)
def _ok() -> None:  # pragma: no cover
    _clean()
    _install(dependency_groups=["dev"])
    _format()
    _lint()
    _test(keepitonehundred=False)
    echo("it's ok!")


@app.command("bump-version", help="Bump the version.")
def _bump_version(
    part: Optional[BumpType] = Option(
        None,
        "-p",
        "--part",
        help="The part of the version to bump. Defaults to None.",
    ),
) -> None:  # pragma: no cover
    project_name = _get_project_name()
    _version = importlib.import_module(f"{project_name}").__version__

    print(f"Current version: {_version}")
    new_version = _bump_version_string(_version, part)
    version_file = f"{project_name}/__init__.py"
    with open(version_file, "r") as f:
        lines = f.readlines()
    with open(version_file, "w") as f:
        for line in lines:
            if line.startswith("__version__"):
                f.write(f'__version__ = "{new_version}"\n')
            else:
                f.write(line)
    print(f"New version: {new_version}")


@app.command(
    "g",
    hidden=True,
    help="Generate fully tested models and routers.",
    no_args_is_help=True,
)
@app.command(
    "generate",
    help="Generate fully tested models and routers.",
    no_args_is_help=True,
)
def _generate(
    content_type: ContentType = Argument(
        ...,
        help="The thing to generate.",
    ),
    _input: List[str] = Argument(
        ...,
    ),
) -> None:
    content_generator_dispatcher = {
        ContentType.model: _ModelContentGenerator,
        ContentType.router: _RouterContentGenerator,
        ContentType.scaffold: _ScaffoldContentGenerator,
    }
    content_generator = content_generator_dispatcher[content_type]()
    content_generator.generate(_input=_input)


@app.command(
    "s",
    hidden=True,
)
@app.command(
    "server",
    help="Run the local dev server.",
)
def _server(
    host: str = Option(
        "127.0.0.1",
        "--host",
        "-h",
        help="The host to run the server on.",
    ),
    port: int = Option(
        8000,
        "--port",
        "-p",
        help="The port to run the server on.",
    ),
    styles: bool = Option(
        False,
        "--styles",
        "-s",
        help="Whether to watch styles.",
    ),
) -> None:  # pragma: no cover
    if styles:
        styles_process = Process(
            name="snok_npx_tailwindcss",
            target=_run_cmd,
            kwargs={
                "cmd": [
                    "npx",
                    "tailwindcss",
                    "-i",
                    "./static/css/input.css",
                    "-o",
                    "./static/css/tailwind.css",
                    "--watch",
                ]
            },
            daemon=True,
        )
        styles_process.start()

    _run_cmd(
        [
            "uvicorn",
            f"{_get_project_name()}.app:app",
            "--host",
            host,
            "--port",
            str(port),
            "--reload",
            "--reload-include",
            "*.html",
            "--reload-include",
            "*.css",
        ]
    )

    if styles:
        styles_process.join()


@app.command("deploy")
def _deploy() -> None:  # pragma: no cover
    """Deploy your code."""
    _run_cmd(
        [
            "ENV=prod",
            "modal",
            "deploy",
            f"{_get_project_name()}/_modal.py",
        ],
        echo=False,
    )



from enum import Enum, unique

APP_NAME = "snok"
APP_DESC = "🚀 A simple, modern, full-stack toolkit for Python 🐍"
PYPROJECT_TOML_FILENAME = "pyproject.toml"
BASE_DEPS = [
    "alembic",
    "asyncpg",
    "fastapi<0.100.0",  # TODO: due to sqlmodel
    "httpx",
    "itsdangerous",
    "jinja2",
    "modal",
    "pydantic<2.0.0",  # TODO: due to sqlmodel
    "python-dateutil",
    "python-dotenv",
    "python-multipart",
    "rich",
    "sqlmodel",
    "structlog",
    "uvicorn",
]

BASE_DEV_DEPS = [
    "asgi-lifespan",
    "black",
    "coverage",
    "greenlet",
    "invoke",
    "mypy",
    "pre-commit",
    "pytest-asyncio",
    "pytest-cov",
    "pytest-xdist",
    "pytest",
    "ruff",
    "watchfiles",
]


@unique
class ProjectType(str, Enum):
    package = "package"
    app = "app"


@unique
class DepencencyAction(str, Enum):
    add = "add"
    remove = "remove"


@unique
class BumpType(Enum):
    MAJOR = "major"
    MINOR = "minor"


@unique
class ContentType(str, Enum):
    model = "model"
    router = "router"
    # TODO: Add support for htmx and tailwind
    # view = "view"
    scaffold = "scaffold"


@unique
class DatabaseCommand(str, Enum):
    migrate = "migrate"
    upgrade = "upgrade"
    downgrade = "downgrade"
    revision = "revision"


@unique
class DeploymentPlatform(str, Enum):
    modal = "modal"



from uuid import uuid4

from httpx import AsyncClient
from {{ __template_name }}.const import API_V0


async def test_create_{{ __template_plural_namespace }}(client: AsyncClient) -> None:
    data = {"name": "John Doe", "age": 30}
    response = await client.post(f"{API_V0}/{{ __template_plural_namespace }}", json=data)
    assert response.status_code == 200, response.text
    assert response.json()["name"] == "John Doe"
    assert response.json()["age"] == 30
    assert response.json()["id"]


async def test_get_{{ __template_plural_namespace }}(client: AsyncClient) -> None:
    data = {"name": "John Doe", "age": 30}
    response = await client.post(f"{API_V0}/{{ __template_plural_namespace }}", json=data)
    assert response.status_code == 200, response.text
    _id = response.json()["id"]
    response = await client.get(f"{API_V0}/{{ __template_plural_namespace }}/{_id}")
    assert response.status_code == 200, response.text
    assert response.json()["name"] == "John Doe"
    assert response.json()["age"] == 30
    assert response.json()["id"]


async def test_update_{{ __template_plural_namespace }}(client: AsyncClient) -> None:
    data = {"name": "John Doe", "age": 30}
    response = await client.post(f"{API_V0}/{{ __template_plural_namespace }}", json=data)
    assert response.status_code == 200, response.text
    _id = response.json()["id"]
    data = {"name": "Jane Doe", "age": 40}
    response = await client.put(f"{API_V0}/{{ __template_plural_namespace }}/{_id}", json=data)
    assert response.status_code == 200, response.text
    assert response.json()["name"] == "Jane Doe"
    assert response.json()["age"] == 40
    assert response.json()["id"]


async def test_delete_{{ __template_plural_namespace }}(client: AsyncClient) -> None:
    data = {"name": "John Doe", "age": 30}
    response = await client.post(f"{API_V0}/{{ __template_plural_namespace }}", json=data)
    assert response.status_code == 200, response.text
    _id = response.json()["id"]
    response = await client.delete(f"{API_V0}/{{ __template_plural_namespace }}/{_id}")
    assert response.status_code == 204
    response = await client.get(
        f"{API_V0}/{{ __template_plural_namespace }}/{_id}", headers={"accept": "application/json"}
    )
    assert response.status_code == 404


async def test_list_{{ __template_plural_namespace }}(client: AsyncClient) -> None:
    data = {"name": "John Doe", "age": 30}
    response = await client.post(f"{API_V0}/{{ __template_plural_namespace }}", json=data)
    assert response.status_code == 200, response.text
    data = {"name": "Jane Doe", "age": 40}
    response = await client.post(f"{API_V0}/{{ __template_plural_namespace }}", json=data)
    assert response.status_code == 200, response.text
    response = await client.get(f"{API_V0}/{{ __template_plural_namespace }}")
    assert response.status_code == 200, response.text
    assert len(response.json()) == 2
    assert response.json()[1]["name"] == "John Doe"
    assert response.json()[1]["age"] == 30
    assert response.json()[0]["name"] == "Jane Doe"
    assert response.json()[0]["age"] == 40
    assert response.json()[0]["id"]
    assert response.json()[1]["id"]


async def test_get_{{ __template_plural_namespace }}_not_found(client: AsyncClient) -> None:
    response = await client.get(f"{API_V0}/{{ __template_plural_namespace }}/{uuid4()}")
    assert response.status_code == 404, response.text
    assert response.json()["detail"] == "Not found"


async def test_update_{{ __template_plural_namespace }}_not_found(client: AsyncClient) -> None:
    data = {"name": "Jane Doe", "age": 40}
    response = await client.put(f"{API_V0}/{{ __template_plural_namespace }}/{uuid4()}", json=data)
    assert response.status_code == 404, response.text
    assert response.json()["detail"] == "Not found"


async def test_delete_{{ __template_plural_namespace }}_not_found(client: AsyncClient) -> None:
    response = await client.delete(f"{API_V0}/{{ __template_plural_namespace }}/{uuid4()}")
    assert response.status_code == 404, response.text
    assert response.json()["detail"] == "Not found"



from typing import Any, List, Optional

from pydantic import UUID4
from sqlmodel import select
from sqlmodel.ext.asyncio.session import AsyncSession

from {{ __template_name }}.models.{{ __template_plural_namespace }} import {{ __template_plural_namespace_caps }}
from {{ __template_name }}.{{ __template_plural_namespace }}.schemas import {{ __template_plural_namespace_caps }}Create, {{ __template_plural_namespace_caps }}QueryParams, {{ __template_plural_namespace_caps }}Update


class {{ __template_plural_namespace_caps }}Service:
    async def create(self, db: AsyncSession, {{ __template_plural_namespace }}_create: {{ __template_plural_namespace_caps }}Create) -> {{ __template_plural_namespace_caps }}:
        db_{{ __template_plural_namespace }} = {{ __template_plural_namespace_caps }}(**{{ __template_plural_namespace }}_create.dict())
        db.add(db_{{ __template_plural_namespace }})
        await db.commit()
        await db.refresh(db_{{ __template_plural_namespace }})
        return db_{{ __template_plural_namespace }}

    async def get(self, db: AsyncSession, _id: UUID4) -> Optional[{{ __template_plural_namespace_caps }}]:
        results = await db.execute(select({{ __template_plural_namespace_caps }}).where({{ __template_plural_namespace_caps }}.id == _id))
        return results.scalars().first()

    async def list(
        self,
        db: AsyncSession,
        params: {{ __template_plural_namespace_caps }}QueryParams,
        order_by: str = "updated_at",
        order_direction: str = "desc",
        offset: int = 0,
        limit: int = 10,
    ) -> List[{{ __template_plural_namespace_caps }}]:
        results = await db.execute(
            select({{ __template_plural_namespace_caps }})
            .filter_by(**params.dict(exclude_none=True))
            .order_by(getattr(getattr({{ __template_plural_namespace_caps }}, order_by), order_direction)())
            .offset(offset)
            .limit(limit)
        )
        return results.scalars().all()

    async def update(
        self, db: AsyncSession, {{ __template_namespace }}: {{ __template_plural_namespace_caps }}, {{ __template_plural_namespace }}_update: {{ __template_plural_namespace_caps }}Update
    ) -> {{ __template_plural_namespace_caps }}:
        update_data = {{ __template_plural_namespace }}_update.dict(exclude_unset=True)
        for key, value in update_data.items():
            setattr({{ __template_namespace }}, key, value)
        db.add({{ __template_namespace }})
        await db.commit()
        await db.refresh({{ __template_namespace }})
        return {{ __template_namespace }}

    async def delete(self, db: AsyncSession, {{ __template_namespace }}: {{ __template_plural_namespace_caps }}) -> None:
        await db.delete({{ __template_namespace }})
        await db.commit()
        return None



import os
import time
from typing import Callable

from fastapi import FastAPI, Request, Response
from fastapi.middleware.cors import CORSMiddleware
from fastapi.routing import APIRoute
from fastapi.staticfiles import StaticFiles
from starlette.middleware.sessions import SessionMiddleware

from {{ __template_name }} import __version__
from {{ __template_name }}.const import LOCAL_DOMAINS, ORIGINS, EXTRA_ORIGINS
from {{ __template_name }}.exceptions import exception_handlers
from {{ __template_name }}.logger import log
from {{ __template_name }}.config import settings
from {{ __template_name }}.health.router import router as health_router
from {{ __template_name }}.router import router

os.environ["TZ"] = "UTC"


def configure_cors(app: FastAPI) -> None:
    if any(domain in settings.APP_URL for domain in LOCAL_DOMAINS):
        ORIGINS.extend(EXTRA_ORIGINS)

    app.add_middleware(
        CORSMiddleware,
        allow_origins=ORIGINS,
        allow_credentials=True,
        allow_methods=["*"],
        allow_headers=["*"],
    )


def add_session_middleware(app: FastAPI) -> None:
    app.add_middleware(
        SessionMiddleware,
        secret_key=settings.API_SECRET_KEY,
    )

# TODO: frontend with htmx and tailwind
# def mount_directories(app: FastAPI) -> None:
#     app.mount(
#         name="static",
#         path="/static",
#         app=StaticFiles(directory="static"),
#     )

def add_http_middleware(app: FastAPI) -> None:
    @app.middleware("http")
    async def add_process_time_header(request: Request, call_next: Callable,) -> Response:
        start_time = time.time() * 1000
        response = await call_next(request)
        request_time_ms = int(time.time() * 1000 - start_time)
        response.headers["x-request-time-ms"] = str(request_time_ms)
        return response


def configure_events(app: FastAPI) -> None:
    @app.on_event("startup")
    async def startup() -> None:
        log.debug({"message": "starting up"})

    @app.on_event("shutdown")
    async def shutdown() -> None:
        log.debug({"message": "shutting down"})


def generate_unique_openapi_id(route: APIRoute) -> str:
    return f"{route.tags[0]}:{route.name}"


def create_app() -> FastAPI:
    app = FastAPI(
        title="{{ __template_name }}",
        version=__version__,
        exception_handlers=exception_handlers,
        generate_unique_id_function=generate_unique_openapi_id
    )

    # TODO: frontend with htmx and tailwind
    # mount_directories(app)
    configure_events(app)
    configure_cors(app)
    add_http_middleware(app)
    add_session_middleware(app)

    app.include_router(health_router)
    app.include_router(router)

    return app


app = create_app()



from packaging.requirements import Requirement

from snok.utils import _get_reqs_from_pyproject_toml


def test_get_reqs_from_pyproject_toml() -> None:
    pyproject_toml = {"project": {"dependencies": ["pendulum"]}}
    result = _get_reqs_from_pyproject_toml(
        pyproject_toml=pyproject_toml,
    )
    assert result == {"pendulum": Requirement("pendulum")}

    pyproject_toml_optional = {
        "project": {
            "dependencies": ["fastapi"],
            "optional-dependencies": {
                "dev": ["pendulum"],
            },
        },
    }
    result = _get_reqs_from_pyproject_toml(
        pyproject_toml=pyproject_toml_optional,
        dependency_group_name="dev",
    )
    assert result == {"pendulum": Requirement("pendulum")}



import json
from typing import Dict, List, Optional

from pydantic import BaseModel, Json, StrictInt, StrictStr, root_validator


class Scope(BaseModel):
    type: StrictStr
    http_version: StrictStr
    method: StrictStr
    scheme: StrictStr
    path: StrictStr
    headers: List
    query_string: bytes


class RequestLoggerMessage(BaseModel):
    scope: Scope
    form: Optional[Dict]
    params: Optional[Dict]
    body: Json | Dict | List | bytes | None


class ResponseLoggerMessage(BaseModel):
    status_code: StrictInt
    raw_headers: List
    body: bytes | Dict | List | None

    @root_validator(pre=True)
    def _validate_body_as_json(cls, values: Dict) -> Dict:
        _headers = values.get("raw_headers", [])
        for header in _headers:
            if header[0].decode("utf8") == "content-type":
                values["body"] = (json.loads(values["body"]) if header[1].decode("utf-8") == "application/json" else None)
        return values



import os
import sys
from typing import AsyncGenerator

from sqlalchemy.ext.asyncio import create_async_engine
from sqlalchemy.orm import sessionmaker
from sqlmodel.ext.asyncio.session import AsyncSession

from {{ __template_name }}.config import settings


def _override_db_url_for_distributed_testing() -> None:
    if os.getenv("_", "").endswith("pytest") or "pytest" in "".join(sys.argv):
        pid = os.getpid()
        settings.POSTGRES_DATABASE = f"{settings.POSTGRES_DATABASE}_{pid}"


_override_db_url_for_distributed_testing()


async_db_engine = create_async_engine(
    url=settings.db_url,
    pool_size=settings.DB_POOL_SIZE,
    max_overflow=settings.DB_MAX_OVERFLOW,
    pool_pre_ping=settings.DB_POOL_PRE_PING,
    future=True,
)


async def db_session() -> AsyncGenerator:
    async_session = sessionmaker(
        async_db_engine, class_=AsyncSession, expire_on_commit=False
    )
    async with async_session() as session:
        yield session



from datetime import datetime
from typing import Optional
from uuid import uuid4

from pydantic import UUID4, BaseModel
from sqlalchemy import Column, DateTime
from sqlmodel import Field, MetaData, SQLModel

_metadata = MetaData()
_metadata.naming_convention = {
    "ix": "ix_%(column_0_label)s",
    "uq": "uq_%(table_name)s_%(column_0_name)s",
    "ck": "ck_%(table_name)s_%(constraint_name)s",
    "fk": "fk_%(table_name)s_%(column_0_name)" "s_%(referred_table_name)s",
    "pk": "pk_%(table_name)s",
}


class TimestampsMixin(BaseModel):
    created_at: Optional[datetime] = Field(
        sa_column=Column(
            DateTime,
            default=datetime.utcnow,
            nullable=False,
        )
    )
    updated_at: Optional[datetime] = Field(
        sa_column=Column(
            DateTime,
            default=datetime.utcnow,
            onupdate=datetime.utcnow,
            nullable=False,
        )
    )
    deleted_at: Optional[datetime] = Field(
        sa_column=Column(
            DateTime,
            nullable=True,
        )
    )


class UUIDMixin(BaseModel):
    id: UUID4 = Field(
        default_factory=uuid4,
        primary_key=True,
        index=True,
        nullable=False,
    )



class Model(SQLModel):
    __abstract__ = True

    metadata = _metadata


class TimestampedModel(Model, TimestampsMixin):
    __abstract__ = True


class RecordModel(TimestampedModel, UUIDMixin):
    __abstract__ = True



from datetime import datetime

from fastapi import APIRouter

from {{ __template_name }} import __version__
from {{ __template_name }}.health.schemas import LivezResponse, ReadyzResponse
from {{ __template_name }}.kit.routers import _APIRoute

router = APIRouter(
    route_class=_APIRoute,
    tags=["health"],
)


@router.get("/livez", response_model=LivezResponse)
async def _livez() -> LivezResponse:
    return LivezResponse(
        message="ok",
    )


@router.get("/readyz", response_model=ReadyzResponse)
async def _readyz() -> ReadyzResponse:
    return ReadyzResponse(
        message="ok",
        version=__version__,
        time=datetime.utcnow(),
    )



from enum import Enum, unique
from pathlib import Path
from typing import List

PARENT = Path(__file__).parent
STATIC_PATH = PARENT / "static"
TEMPLATES_PATH = PARENT / "templates"
ORIGINS: List = []
LOCAL_DOMAINS: List = ["localhost","127.0.0.1"]
EXTRA_ORIGINS = ["http://localhost", "http://127.0.0.1", "http://localhost:8000", "http://127.0.0.1:8000"]
API_V0 = "/api/v0"

@unique
class ENV(str, Enum):
    dev = "dev"
    test = "test"
    prod = "prod"



import os

from httpx import AsyncClient

from {{ __template_name }} import __version__
from {{ __template_name }}.config import settings


async def test_tz() -> None:
    assert os.environ["TZ"] == "UTC"


async def test_livez(client: AsyncClient) -> None:
    response = await client.get("/livez")
    assert response.status_code == 200
    assert response.headers["content-type"] == "application/json"
    assert response.json()["message"] == "ok"
    assert "version" not in response.json()


async def test_readyz(client: AsyncClient) -> None:
    response = await client.get("/readyz")
    assert response.status_code == 200
    assert response.headers["content-type"] == "application/json"
    assert response.json()["message"] == "ok"
    assert response.json()["version"] == __version__


async def test_settings() -> None:
    assert settings.ENV == "test"



from fastapi import APIRouter

from {{ __template_name }}.kit.routers import _APIRoute

router = APIRouter(
    route_class=_APIRoute,
    tags=["{{ __template_router_name }}"],
)

{% for route in __template_router_routes %}
@router.get("/{{ route }}", response_model=None)
async def _{{ route }}() -> None:
    return None
{% endfor %}



import importlib.resources as resources
import os
import re

from jinja2 import Template

from snok.const import ProjectType


class BaseNewService:
    def __init__(self) -> None:
        import snok

        self.root_template_dir = str(resources.files(snok)) + "/templates"

    def create(
        self,
        name: str,
        desc: str,
        output_dir: str,
        type: ProjectType = ProjectType.package,
    ) -> None:
        raise NotImplementedError("create() must be implemented by subclass.")

    def _render_content_directory(
        self,
        name: str,
        desc: str,
        source_dir: str,
        output_dir: str,
        type: ProjectType = ProjectType.package,
    ) -> None:
        for root, dirs, files in os.walk(source_dir):
            if not ("__pycache__" in root or "_cache" in root):
                for file in files:
                    file_path = os.path.join(root, file)
                    relative_path = os.path.relpath(file_path, source_dir)
                    output_path = os.path.join(output_dir, relative_path)

                    if output_path.endswith("_pyproject_toml"):
                        output_path = output_path.replace(
                            "_pyproject_toml", "pyproject.toml"
                        )
                        if type == ProjectType.app:
                            with open(file_path, "r") as f:
                                content = f.read()
                                if "--asyncio-mode=auto" not in content:
                                    content = content.replace(
                                        '"console_output_style=progress"',
                                        '"console_output_style=progress"\n'
                                        ', "--asyncio-mode=auto"',
                                    )
                            tmp_file = file_path + ".tmp"
                            with open(tmp_file, "w") as f:
                                f.write(content)
                            self._render_content_file(
                                name=name,
                                desc=desc,
                                source_path=tmp_file,
                                output_path=output_path,
                            )
                            os.remove(tmp_file)
                            continue

                    if output_path.endswith("alembic.ini"):
                        output_path = output_path.replace(
                            "migrations/alembic.ini", "alembic.ini"
                        )

                    if output_path.endswith("py.typed"):
                        output_path = output_path.replace(
                            "py.typed", f"{name}/py.typed"
                        )

                    # TODO: Add support for htmx and tailwind
                    if output_path.endswith("tailwind.config.js"):
                        output_path = output_path.replace(  # pragma: no cover
                            "static/tailwind.config.js", "tailwind.config.js"
                        )

                    _basename = os.path.basename(relative_path)
                    if _basename.startswith("_.") and not _basename.endswith(".py"):
                        _newbasename = _basename.replace("_.", ".")
                        output_path = os.path.join(
                            output_dir,
                            relative_path.replace(_basename, _newbasename),
                        )

                    if re.match(r"^.*\.env.*$", output_path):
                        b = os.path.basename(output_path)
                        output_path = output_path.replace(f"{name}/{b}", b)

                    self._render_content_file(
                        name=name,
                        desc=desc,
                        source_path=file_path,
                        output_path=output_path,
                    )

    def _render_content_file(
        self,
        name: str,
        desc: str,
        source_path: str,
        output_path: str,
    ) -> None:
        os.makedirs(os.path.dirname(output_path), exist_ok=True)
        if self._file_is_image_file(file_path=source_path):
            # TODO: Add support for images for rendering in frontend
            with open(source_path, "rb") as source_file:  # pragma: no cover
                with open(output_path, "wb") as output_file:
                    output_file.write(source_file.read())
        else:
            with open(source_path, "r") as source_file:
                source_content = source_file.read()
                with open(output_path, "w") as output_file:
                    output_file.write(
                        self._render_output_content_string(
                            source_path=source_path,
                            source_content=source_content,
                            name=name,
                            desc=desc,
                        )
                    )

    def _file_is_image_file(self, file_path: str) -> bool:
        return os.path.splitext(file_path)[1] in [".png", ".jpg", ".jpeg", ".gif"]

    def _render_output_content_string(
        self, source_path: str, source_content: str, name: str, desc: str
    ) -> str:
        return (
            source_content
            if os.path.splitext(source_path)[1] == ".html"
            else Template(source_content).render(
                __template_name=name,
                __template_description=desc,
            )
        )


class NewPackageService(BaseNewService):
    def __init__(self) -> None:
        super().__init__()

    def create(
        self,
        name: str,
        desc: str,
        output_dir: str,
        type: ProjectType = ProjectType.package,
    ) -> None:
        shared_template_dir = f"{self.root_template_dir}/__shared"
        package_template_dir = f"{self.root_template_dir}/__{type.value}"
        tests_template_dir = f"{self.root_template_dir}/__{type.value}_tests"
        os.makedirs(output_dir, exist_ok=True)
        self._render_content_directory(
            name=name,
            desc=desc,
            source_dir=shared_template_dir,
            output_dir=output_dir,
            type=type,
        )
        package_root = f"{output_dir}/{name}"
        os.makedirs(package_root, exist_ok=True)
        self._render_content_directory(
            name=name,
            desc=desc,
            source_dir=package_template_dir,
            output_dir=package_root,
            type=type,
        )
        package_test_root = f"{output_dir}/tests"
        os.makedirs(package_test_root, exist_ok=True)
        self._render_content_directory(
            name=name,
            desc=desc,
            source_dir=tests_template_dir,
            output_dir=package_test_root,
            type=type,
        )
        if not os.path.exists(f"{output_dir}/.git"):
            os.system(f"git init {output_dir} > /dev/null 2>&1")


class NewAppService(BaseNewService):
    def __init__(self) -> None:
        super().__init__()

    def create(
        self,
        name: str,
        desc: str,
        output_dir: str,
        type: ProjectType = ProjectType.app,
    ) -> None:
        shared_template_dir = f"{self.root_template_dir}/__shared"
        app_template_dir = f"{self.root_template_dir}/__{type.value}"
        app_tests_template_dir = f"{self.root_template_dir}/__{type.value}_tests"
        app_migrations_templates_dir = (
            f"{self.root_template_dir}/__{type.value}_migrations"
        )
        # TODO: frontend with htmx and tailwind
        # app_views_templates_dir = f"{self.root_template_dir}/__{type.value}_views"
        # app_static_templates_dir = f"{self.root_template_dir}/__{type.value}_static"
        os.makedirs(output_dir, exist_ok=True)
        self._render_content_directory(
            name=name,
            desc=desc,
            source_dir=shared_template_dir,
            output_dir=output_dir,
            type=type,
        )
        app_root = f"{output_dir}/{name}"
        os.makedirs(app_root, exist_ok=True)
        self._render_content_directory(
            name=name,
            desc=desc,
            source_dir=app_template_dir,
            output_dir=app_root,
            type=type,
        )
        app_test_root = f"{output_dir}/tests"
        os.makedirs(app_test_root, exist_ok=True)
        self._render_content_directory(
            name=name,
            desc=desc,
            source_dir=app_tests_template_dir,
            output_dir=app_test_root,
            type=type,
        )
        app_migrations_root = f"{output_dir}/migrations"
        os.makedirs(app_migrations_root, exist_ok=True)
        self._render_content_directory(
            name=name,
            desc=desc,
            source_dir=app_migrations_templates_dir,
            output_dir=app_migrations_root,
            type=type,
        )
        # TODO: frontend with htmx and tailwind
        # app_views_root = f"{output_dir}/views"
        # os.makedirs(app_views_root, exist_ok=True)
        # self._render_content_directory(
        #     name=name,
        #     desc=desc,
        #     source_dir=app_views_templates_dir,
        #     output_dir=app_views_root,
        #     type=type,
        # )
        # app_template_root = f"{output_dir}/static"
        # os.makedirs(app_template_root, exist_ok=True)
        # self._render_content_directory(
        #     name=name,
        #     desc=desc,
        #     source_dir=app_static_templates_dir,
        #     output_dir=app_template_root,
        #     type=type,
        # )
        if not os.path.exists(f"{output_dir}/.git"):
            os.system(f"git init {output_dir} > /dev/null 2>&1")



import asyncio
import os
import time
from logging.config import fileConfig

from alembic import context
from sqlalchemy import create_engine, pool
from sqlalchemy.engine.base import Connection
from sqlmodel.ext.asyncio.session import AsyncEngine

from {{ __template_name }}.config import settings
from {{ __template_name }}.models import RecordModel

# configure local env setup
os.environ["TZ"] = "UTC"
tz = time.strftime("%z")

# this is the Alembic Config object, which provides
# access to the values within the .ini file in use.
config = context.config

# Interpret the config file for Python logging.
# This line sets up loggers basically.
if config.config_file_name is not None:
    fileConfig(config.config_file_name)

# add your model's MetaData object here
# for 'autogenerate' support
target_metadata = RecordModel.metadata

# other values from the config, defined by the needs of env.py,
# can be acquired:
# my_important_option = config.get_main_option("my_important_option")
# ... etc.


def get_url() -> str:
    url = settings.db_url
    if not url:
        raise ValueError("database url is not set")
    return url


def run_migrations_offline() -> None:
    """Run migrations in 'offline' mode.

    This configures the context with just a URL
    and not an Engine, though an Engine is acceptable
    here as well.  By skipping the Engine creation
    we don't even need a DBAPI to be available.

    Calls to context.execute() here emit the given string to the
    script output.

    """
    url = get_url()
    context.configure(
        url=url,
        target_metadata=target_metadata,
        literal_binds=True,
        dialect_opts={"paramstyle": "named"},
    )

    with context.begin_transaction():
        context.run_migrations()


def _run_migrations_sync(connection: Connection) -> None:
    context.configure(connection=connection, target_metadata=target_metadata)

    with context.begin_transaction():
        context.run_migrations()


async def run_migrations_online() -> None:
    """Run migrations in 'online' mode.

    In this scenario we need to create an Engine
    and associate a connection with the context.
    """
    connectable = AsyncEngine(
        create_engine(get_url(), future=True, poolclass=pool.NullPool)
    )

    async with connectable.connect() as connection:
        await connection.run_sync(_run_migrations_sync)

    await connectable.dispose()


if context.is_offline_mode():
    run_migrations_offline()
else:
    asyncio.run(run_migrations_online())



import os
from typing import Any, List

from jinja2 import Template

from snok.utils import _get_project_name, _get_snok_path


class _BaseContentGenerator:
    def generate(self, *args: Any, **kwargs: Any) -> Any:
        raise NotImplementedError("Subclasses must implement this method.")

    def _pluralize_name(self, name: str) -> str:
        if name.endswith(("s", "x", "z", "ch", "sh")):
            return name + "es"
        elif name.endswith("y"):
            if name[-2] in ("a", "e", "i", "o", "u"):
                return name + "s"
            else:
                return name[:-1] + "ies"
        else:
            return name + "s"


class _ModelContentGenerator(_BaseContentGenerator):
    MISSING_INPUT_ERROR = (
        "Missing input. Please provide a model name and fields."
        " For example: snok generate model user name:str email:str"
    )
    BASE_FIELD_TYPES = ["str", "int", "float", "bool", "datetime"]

    def _validate_field_types(self, fields: Any) -> None:
        field_set = set(fields)
        try:
            field_types_set = set([field.split(":")[1].lower() for field in field_set])
        except IndexError:
            raise ValueError(
                "Invalid field format. Please provide a field name and type."
                " For example: snok generate model user name:str email:str"
            )
        if not field_types_set.issubset(self.BASE_FIELD_TYPES):
            invalid_field_types = field_types_set - set(self.BASE_FIELD_TYPES)
            raise ValueError(
                f"Invalid field types: {','.join(invalid_field_types)}"
                f" Please provide a valid field type, "
                f"such as {','.join(self.BASE_FIELD_TYPES)}."
            )

    def generate(self, *args: Any, **kwargs: Any) -> Any:
        _input = kwargs.get("_input")
        assert _input is not None, self.MISSING_INPUT_ERROR
        namespace, fields = _input[0].lower(), _input[1:]
        _plural_namespace = self._pluralize_name(namespace)
        self._validate_field_types(fields)
        _fields = [field.split(":") for field in fields]
        self._write_model_to_models(
            namespace=namespace, plural_namespace=_plural_namespace, fields=_fields
        )
        self._update_models_init_file(plural_namespace=_plural_namespace)

    def _write_model_to_models(
        self, namespace: str, plural_namespace: str, fields: List[List[str]]
    ) -> None:
        models_dir = f"{_get_project_name()}/models"
        model_filename = f"{models_dir}/{plural_namespace}.py"
        model_template_file = _get_snok_path() + "/templates/__app_models/model.py"
        content = Template(open(model_template_file).read()).render(
            __template_name=_get_project_name(),
            __template_namespace=namespace,
            __template_plural_namespace=plural_namespace,
            __template_plural_namespace_caps=plural_namespace.capitalize(),
            __template_fields=fields,
        )
        with open(model_filename, "w") as f:
            f.write(content)

    def _update_models_init_file(self, plural_namespace: str) -> None:
        models_dir = f"{_get_project_name()}/models"
        init_filename = f"{models_dir}/__init__.py"
        import_statement = (
            f"from {_get_project_name()}.models.{plural_namespace}"
            f" import {plural_namespace.capitalize()}\n"
        )
        with open(init_filename, "r") as f:
            lines = f.readlines()
        with open(init_filename, "w") as f:
            for line in lines:
                if line.startswith("__all__"):
                    f.write(line.replace("__all__", f"{import_statement}\n__all__"))
                elif line.startswith("]"):
                    f.write(line.replace("]", f'"{plural_namespace.capitalize()}",\n]'))
                else:
                    f.write(line)


class _RouterContentGenerator(_BaseContentGenerator):
    def generate(self, *args: Any, **kwargs: Any) -> Any:
        _input = kwargs.get("_input")
        assert _input is not None, (
            "Missing input. Please provide a router name and routes."
            " For example: snok generate router myrouter route1 route2 route3"
        )
        namespace, routes = _input[0].lower(), _input[1:]
        os.makedirs(f"{_get_project_name()}/{namespace}", exist_ok=True)
        with open(f"{_get_project_name()}/{namespace}/__init__.py", "w") as f:
            pass
        router_filename = f"{_get_project_name()}/{namespace}/router.py"
        router_template_file = _get_snok_path() + "/templates/__app_router/router.py"
        content = Template(open(router_template_file).read()).render(
            __template_name=_get_project_name(),
            __template_router_name=namespace,
            __template_router_routes=routes,
        )
        with open(router_filename, "w") as router_py:
            router_py.write(content)

        # update routers.py with import statement
        routers_file = f"{_get_project_name()}/router.py"
        import_statement = (
            f"from {_get_project_name()}.{namespace}.router"
            f" import router as {namespace}_router\n"
        )
        with open(routers_file, "r") as f:
            lines = f.readlines()
        with open(routers_file, "w") as f:
            for line in lines:
                if line.startswith("router = APIRouter(\n"):
                    f.write(
                        line.replace(
                            "router = APIRouter(\n",
                            f"{import_statement}\n\nrouter = APIRouter(\n",
                        )
                    )
                else:
                    f.write(line)
            f.write(f"\nrouter.include_router(router={namespace}_router)\n")


class _ScaffoldContentGenerator(_BaseContentGenerator):
    def generate(self, *args: Any, **kwargs: Any) -> Any:
        _input = kwargs.get("_input")
        assert _input is not None, (
            "Missing input. Please provide a model name and fields."
            " For example: snok generate scaffold person name:str age:int"
        )
        mcg = _ModelContentGenerator()
        mcg.generate(*args, **kwargs)

        namespace, fields = _input[0].lower(), _input[1:]
        _fields = [field.split(":") for field in fields]
        _plural_namespace = self._pluralize_name(namespace)
        os.makedirs(f"{_get_project_name()}/{_plural_namespace}", exist_ok=True)
        with open(f"{_get_project_name()}/{_plural_namespace}/__init__.py", "w") as f:
            pass
        os.makedirs(
            f"{_get_project_name()}/../tests/{_plural_namespace}", exist_ok=True
        )
        with open(
            f"{_get_project_name()}/../tests/{_plural_namespace}/__init__.py", "w"
        ) as f:
            pass

        # generate scaffolded router
        scaffolded_router_filename_dst = (
            f"{_get_project_name()}/{_plural_namespace}/router.py"
        )
        scaffolded_router_filename_src = (
            f"{_get_snok_path()}/templates/__app_scaffold/router.py"
        )
        content = Template(open(scaffolded_router_filename_src).read()).render(
            __template_name=_get_project_name(),
            __template_plural_namespace=_plural_namespace,
            __template_plural_namespace_caps=_plural_namespace.capitalize(),
            __template_namespace=namespace,
        )
        with open(scaffolded_router_filename_dst, "w") as f:
            f.write(content)

        # generate scaffolded service
        scaffolded_service_filename_dst = (
            f"{_get_project_name()}/{_plural_namespace}/service.py"
        )
        scaffolded_service_filename_src = (
            f"{_get_snok_path()}/templates/__app_scaffold/service.py"
        )
        content = Template(open(scaffolded_service_filename_src).read()).render(
            __template_name=_get_project_name(),
            __template_plural_namespace=_plural_namespace,
            __template_plural_namespace_caps=_plural_namespace.capitalize(),
            __template_namespace=namespace,
        )
        with open(scaffolded_service_filename_dst, "w") as f:
            f.write(content)

        # generate scaffolded schemas
        scaffolded_schemas_filename_dst = (
            f"{_get_project_name()}/{_plural_namespace}/schemas.py"
        )
        scaffolded_schemas_filename_src = (
            f"{_get_snok_path()}/templates/__app_scaffold/schemas.py"
        )
        content = Template(open(scaffolded_schemas_filename_src).read()).render(
            __template_name=_get_project_name(),
            __template_plural_namespace=_plural_namespace,
            __template_plural_namespace_caps=_plural_namespace.capitalize(),
            __template_namespace=namespace,
            __template_fields=_fields,
        )
        with open(scaffolded_schemas_filename_dst, "w") as f:
            f.write(content)

        # generate scaffolded tests

        scaffolded_tests_filename_dst = (
            f"{_get_project_name()}/../tests/{_plural_namespace}/test_router.py"
        )
        scaffolded_tests_filename_src = (
            f"{_get_snok_path()}/templates/__app_scaffold_tests/test_router.py"
        )
        content = Template(open(scaffolded_tests_filename_src).read()).render(
            __template_name=_get_project_name(),
            __template_plural_namespace=_plural_namespace,
            __template_plural_namespace_caps=_plural_namespace.capitalize(),
            __template_namespace=namespace,
        )
        with open(scaffolded_tests_filename_dst, "w") as f:
            f.write(content)

        # update routers.py with import statement
        routers_file = f"{_get_project_name()}/router.py"
        import_statement = (
            f"from {_get_project_name()}.{_plural_namespace}.router"
            f" import router as {_plural_namespace}_router\n"
        )
        with open(routers_file, "r") as f:
            lines = f.readlines()
        with open(routers_file, "w") as f:
            for line in lines:
                if line.startswith("router = APIRouter(\n"):
                    f.write(
                        line.replace(
                            "router = APIRouter(\n",
                            f"{import_statement}\n\nrouter = APIRouter(\n",
                        )
                    )
                else:
                    f.write(line)
            f.write(f"\nrouter.include_router(router={_plural_namespace}_router)\n")



from typing import Optional

from pydantic import BaseModel


class {{ __template_plural_namespace_caps }}QueryParams(BaseModel):
    {% for field in __template_fields %}
    {{ field[0] }}: Optional[{{ field[1] }}] = None{% endfor %}


class {{ __template_plural_namespace_caps }}Base(BaseModel):
    {% for field in __template_fields %}
    {{ field[0] }}: {{ field[1] }}{% endfor %}


class {{ __template_plural_namespace_caps }}Create({{ __template_plural_namespace_caps }}Base):
    ...


class {{ __template_plural_namespace_caps }}Update({{ __template_plural_namespace_caps }}Base):
    ...



from datetime import datetime

from pydantic import BaseModel, StrictStr


class LivezResponse(BaseModel):
    message: StrictStr


class ReadyzResponse(BaseModel):
    message: StrictStr
    version: StrictStr
    time: datetime



import os
from unittest import mock

from typer.testing import CliRunner

from snok import __version__
from snok.cli import app


def test_version(cli_runner: CliRunner) -> None:
    response = cli_runner.invoke(app, ["version"])
    assert response.exit_code == 0
    assert response.stdout == __version__


@mock.patch("snok.cli._add")
def test_new_package_setup(
    mock_add: mock.MagicMock, cli_runner: CliRunner, setup_test_project_dir: str
) -> None:
    response = cli_runner.invoke(
        app,
        ["new", "test_package", "-o", setup_test_project_dir],
    )
    assert response.exit_code == 0

    # assert the output directory exists
    for path in [
        f"{setup_test_project_dir}/test_package",
        f"{setup_test_project_dir}/tests",
        f"{setup_test_project_dir}/.git",
        f"{setup_test_project_dir}/README.md",
        f"{setup_test_project_dir}/pyproject.toml",
        f"{setup_test_project_dir}/.gitignore",
        f"{setup_test_project_dir}/test_package/__init__.py",
        f"{setup_test_project_dir}/test_package/main.py",
        f"{setup_test_project_dir}/tests/__init__.py",
        f"{setup_test_project_dir}/tests/conftest.py",
        f"{setup_test_project_dir}/tests/test_main.py",
        f"{setup_test_project_dir}/tests/test_version.py",
    ]:
        assert os.path.exists(path)


@mock.patch("snok.cli._add")
def test_new_app_setup(
    mock_add: mock.MagicMock, cli_runner: CliRunner, setup_test_project_dir: str
) -> None:
    response = cli_runner.invoke(
        app,
        [
            "new",
            "test_app",
            "-o",
            setup_test_project_dir,
            "--type",
            "app",
        ],
    )
    assert response.exit_code == 0

    # assert the output directory exists
    for path in [
        f"{setup_test_project_dir}/test_app",
        f"{setup_test_project_dir}/tests",
        f"{setup_test_project_dir}/.git",
        f"{setup_test_project_dir}/README.md",
        f"{setup_test_project_dir}/pyproject.toml",
        f"{setup_test_project_dir}/.gitignore",
        f"{setup_test_project_dir}/test_app/__init__.py",
        f"{setup_test_project_dir}/test_app/router.py",
        f"{setup_test_project_dir}/test_app/logger.py",
        f"{setup_test_project_dir}/migrations/env.py",
        f"{setup_test_project_dir}/test_app/kit/db.py",
        f"{setup_test_project_dir}/tests/__init__.py",
        f"{setup_test_project_dir}/tests/conftest.py",
        f"{setup_test_project_dir}/tests/health/__init__.py",
    ]:
        assert os.path.exists(path)


@mock.patch("snok.cli._add")
def test_generate_model(
    mock_add: mock.MagicMock, cli_runner: CliRunner, setup_test_project_dir: str
) -> None:
    response = cli_runner.invoke(
        app,
        [
            "new",
            "test_generate_model_app",
            "-o",
            setup_test_project_dir,
            "--type",
            "app",
        ],
    )
    assert response.exit_code == 0

    cwd = os.getcwd()
    os.chdir(f"{setup_test_project_dir}")
    response = cli_runner.invoke(
        app,
        ["generate", "model", "person", "name:str", "age:int"],
    )
    os.chdir(cwd)
    assert response.exit_code == 0

    # assert the output directory exists
    for path in [
        f"{setup_test_project_dir}/test_generate_model_app",
        f"{setup_test_project_dir}/tests",
        f"{setup_test_project_dir}/.git",
        f"{setup_test_project_dir}/README.md",
        f"{setup_test_project_dir}/pyproject.toml",
        f"{setup_test_project_dir}/.gitignore",
        f"{setup_test_project_dir}/test_generate_model_app/__init__.py",
        f"{setup_test_project_dir}/test_generate_model_app/router.py",
        f"{setup_test_project_dir}/test_generate_model_app/logger.py",
        f"{setup_test_project_dir}/migrations/env.py",
        f"{setup_test_project_dir}/test_generate_model_app/kit/db.py",
        f"{setup_test_project_dir}/tests/__init__.py",
        f"{setup_test_project_dir}/tests/conftest.py",
        f"{setup_test_project_dir}/tests/health/__init__.py",
    ]:
        assert os.path.exists(path)


@mock.patch("snok.cli._add")
def test_generate_model_fails(
    mock_add: mock.MagicMock, cli_runner: CliRunner, setup_test_project_dir: str
) -> None:
    response = cli_runner.invoke(
        app,
        [
            "new",
            "test_generate_model_app",
            "-o",
            setup_test_project_dir,
            "--type",
            "app",
        ],
    )
    assert response.exit_code == 0

    cwd = os.getcwd()
    os.chdir(f"{setup_test_project_dir}")
    response = cli_runner.invoke(
        app,
        ["generate", "model"],
    )
    os.chdir(cwd)
    assert response.exit_code == 2

    cwd = os.getcwd()
    os.chdir(f"{setup_test_project_dir}")
    response = cli_runner.invoke(
        app,
        ["generate", "model", "person", "name:"],
    )
    os.chdir(cwd)
    assert response.exit_code == 1

    cwd = os.getcwd()
    os.chdir(f"{setup_test_project_dir}")
    response = cli_runner.invoke(
        app,
        ["generate", "model", "person", "name"],
    )
    os.chdir(cwd)
    assert response.exit_code == 1


@mock.patch("snok.cli._add")
def test_generate_router(
    mock_add: mock.MagicMock, cli_runner: CliRunner, setup_test_project_dir: str
) -> None:
    response = cli_runner.invoke(
        app,
        [
            "new",
            "test_generate_router_app",
            "-o",
            setup_test_project_dir,
            "--type",
            "app",
        ],
    )
    assert response.exit_code == 0

    cwd = os.getcwd()
    os.chdir(f"{setup_test_project_dir}")
    response = cli_runner.invoke(
        app,
        ["generate", "router", "myrouter", "hello", "world"],
    )
    os.chdir(cwd)
    assert response.exit_code == 0

    # assert the output directory exists
    for path in [
        f"{setup_test_project_dir}/test_generate_router_app",
        f"{setup_test_project_dir}/tests",
        f"{setup_test_project_dir}/.git",
        f"{setup_test_project_dir}/README.md",
        f"{setup_test_project_dir}/pyproject.toml",
        f"{setup_test_project_dir}/.gitignore",
        f"{setup_test_project_dir}/test_generate_router_app/__init__.py",
        f"{setup_test_project_dir}/test_generate_router_app/router.py",
        f"{setup_test_project_dir}/test_generate_router_app/myrouter/router.py",
        f"{setup_test_project_dir}/test_generate_router_app/logger.py",
        f"{setup_test_project_dir}/migrations/env.py",
        f"{setup_test_project_dir}/test_generate_router_app/kit/db.py",
        f"{setup_test_project_dir}/tests/__init__.py",
        f"{setup_test_project_dir}/tests/conftest.py",
        f"{setup_test_project_dir}/tests/health/__init__.py",
    ]:
        assert os.path.exists(path)

    # assert the hello and world functions exist
    for path in [
        f"{setup_test_project_dir}/test_generate_router_app/myrouter/router.py",
    ]:
        assert "async def _hello" in open(path).read()
        assert "async def _world" in open(path).read()


@mock.patch("snok.cli._add")
def test_generate_scaffold(
    mock_add: mock.MagicMock, cli_runner: CliRunner, setup_test_project_dir: str
) -> None:
    response = cli_runner.invoke(
        app,
        [
            "new",
            "test_generate_scaffold_app",
            "-o",
            setup_test_project_dir,
            "--type",
            "app",
        ],
    )
    assert response.exit_code == 0

    cwd = os.getcwd()
    os.chdir(f"{setup_test_project_dir}")
    response = cli_runner.invoke(
        app,
        ["generate", "scaffold", "person", "name:str", "age:int"],
    )
    os.chdir(cwd)
    assert response.exit_code == 0

    # assert the output directory exists
    for path in [
        f"{setup_test_project_dir}/test_generate_scaffold_app",
        f"{setup_test_project_dir}/tests",
        f"{setup_test_project_dir}/.git",
        f"{setup_test_project_dir}/README.md",
        f"{setup_test_project_dir}/pyproject.toml",
        f"{setup_test_project_dir}/.gitignore",
        f"{setup_test_project_dir}/test_generate_scaffold_app/__init__.py",
        f"{setup_test_project_dir}/test_generate_scaffold_app/router.py",
        f"{setup_test_project_dir}/test_generate_scaffold_app/logger.py",
        f"{setup_test_project_dir}/migrations/env.py",
        f"{setup_test_project_dir}/test_generate_scaffold_app/kit/db.py",
        f"{setup_test_project_dir}/test_generate_scaffold_app/persons/router.py",
        f"{setup_test_project_dir}/test_generate_scaffold_app/persons/schemas.py",
        f"{setup_test_project_dir}/test_generate_scaffold_app/persons/service.py",
        f"{setup_test_project_dir}/tests/__init__.py",
        f"{setup_test_project_dir}/tests/conftest.py",
        f"{setup_test_project_dir}/tests/health/__init__.py",
    ]:
        assert os.path.exists(path)



from snok.services.generator import _BaseContentGenerator


def test_pluralize_name() -> None:
    _bcg = _BaseContentGenerator()
    assert _bcg._pluralize_name("user") == "users"
    assert _bcg._pluralize_name("users") == "userses"
    assert _bcg._pluralize_name("User") == "Users"
    assert _bcg._pluralize_name("Users") == "Userses"
    assert _bcg._pluralize_name("ax") == "axes"
    assert _bcg._pluralize_name("bogey") == "bogeys"
    assert _bcg._pluralize_name("policy") == "policies"



from typing import List

from fastapi import APIRouter, Body, Depends, HTTPException, Response, status
from pydantic import UUID4
from sqlmodel.ext.asyncio.session import AsyncSession

from {{ __template_name }}.database import db_session
from {{ __template_name }}.kit.routers import _APIRoute
from {{ __template_name }}.models.{{ __template_plural_namespace }} import {{ __template_plural_namespace_caps }}
from {{ __template_name }}.{{ __template_plural_namespace }}.schemas import {{ __template_plural_namespace_caps }}Create, {{ __template_plural_namespace_caps }}QueryParams, {{ __template_plural_namespace_caps }}Update
from {{ __template_name }}.{{ __template_plural_namespace }}.service import {{ __template_plural_namespace_caps }}Service

router = APIRouter(
    route_class=_APIRoute,
    tags=["{{ __template_plural_namespace }}"],
)


@router.get("/{{ __template_plural_namespace }}", response_model=List[{{ __template_plural_namespace_caps }}])
async def _list_{{ __template_plural_namespace }}(
    params: {{ __template_plural_namespace_caps }}QueryParams = Depends(),
    order_by: str = "updated_at",
    order_direction: str = "desc",
    offset: int = 0,
    limit: int = 10,
    db: AsyncSession = Depends(db_session),
    {{ __template_plural_namespace }}_service: {{ __template_plural_namespace_caps }}Service = Depends({{ __template_plural_namespace_caps }}Service),
) -> List[{{ __template_plural_namespace_caps }}]:
    return await {{ __template_plural_namespace }}_service.list(
        params=params,
        db=db,
        order_by=order_by,
        order_direction=order_direction,
        offset=offset,
        limit=limit,
    )


@router.post("/{{ __template_plural_namespace }}", response_model={{ __template_plural_namespace_caps }})
async def _create_{{ __template_plural_namespace }}(
    {{ __template_plural_namespace }}_create: {{ __template_plural_namespace_caps }}Create = Body(...),
    db: AsyncSession = Depends(db_session),
    {{ __template_plural_namespace }}_service: {{ __template_plural_namespace_caps }}Service = Depends({{ __template_plural_namespace_caps }}Service),
) -> {{ __template_plural_namespace_caps }}:
    return await {{ __template_plural_namespace }}_service.create(db=db, {{ __template_plural_namespace }}_create={{ __template_plural_namespace }}_create)


@router.get("/{{ __template_plural_namespace }}/{_id}", response_model={{ __template_plural_namespace_caps }})
async def _get_{{ __template_plural_namespace }}(
    _id: UUID4,
    db: AsyncSession = Depends(db_session),
    {{ __template_plural_namespace }}_service: {{ __template_plural_namespace_caps }}Service = Depends({{ __template_plural_namespace_caps }}Service),
) -> {{ __template_plural_namespace_caps }}:
    {{ __template_namespace }} = await {{ __template_plural_namespace }}_service.get(db=db, _id=_id)
    if not {{ __template_namespace }}:
        raise HTTPException(status_code=404, detail="Not found")
    return {{ __template_namespace }}


@router.put("/{{ __template_plural_namespace }}/{_id}", response_model={{ __template_plural_namespace_caps }})
async def _update_{{ __template_plural_namespace }}(
    _id: UUID4,
    {{ __template_plural_namespace }}_update: {{ __template_plural_namespace_caps }}Update = Body(...),
    db: AsyncSession = Depends(db_session),
    {{ __template_plural_namespace }}_service: {{ __template_plural_namespace_caps }}Service = Depends({{ __template_plural_namespace_caps }}Service),
) -> {{ __template_plural_namespace_caps }}:
    {{ __template_namespace }} = await {{ __template_plural_namespace }}_service.get(db=db, _id=_id)
    if not {{ __template_namespace }}:
        raise HTTPException(status_code=404, detail="Not found")
    return await {{ __template_plural_namespace }}_service.update(
        db=db, {{ __template_namespace }}={{ __template_namespace }}, {{ __template_plural_namespace }}_update={{ __template_plural_namespace }}_update
    )


@router.delete("/{{ __template_plural_namespace }}/{_id}", response_class=Response)
async def _delete_{{ __template_plural_namespace }}(
    _id: UUID4,
    db: AsyncSession = Depends(db_session),
    {{ __template_plural_namespace }}_service: {{ __template_plural_namespace_caps }}Service = Depends({{ __template_plural_namespace_caps }}Service),
) -> Response:
    {{ __template_namespace }} = await {{ __template_plural_namespace }}_service.get(db=db, _id=_id)
    if not {{ __template_namespace }}:
        raise HTTPException(status_code=404, detail="Not found")
    await {{ __template_plural_namespace }}_service.delete(db=db, {{ __template_namespace }}={{ __template_namespace }})
    return Response(status_code=status.HTTP_204_NO_CONTENT)



import os
from typing import Any, Dict, Iterable, List, Optional

import httpx
import toml
from invoke import Context
from packaging.requirements import Requirement, SpecifierSet
from typer import Exit

from snok.const import PYPROJECT_TOML_FILENAME, BumpType, DepencencyAction


def _get_default_output_dir() -> str:
    """Returns the default output directory of the project."""
    return os.path.dirname(os.getenv("VIRTUAL_ENV", os.getcwd()))


def _get_snok_path() -> str:
    """Returns the path to the snok package."""
    return os.path.dirname(os.path.abspath(__file__))


def _run_cmd(
    cmd: List[str],
    pty: bool = True,
    echo: bool = True,
) -> None:  # pragma: no cover
    ctx = Context()
    try:
        ctx.run(
            " ".join(cmd),
            pty=pty,
            echo=echo,
        )
    except Exception:
        raise Exit(1)


def _get_package_version_from_pypi(r: Requirement) -> SpecifierSet:  # pragma: no cover
    response = httpx.get(f"https://pypi.org/pypi/{r.name}/json")
    return SpecifierSet(f'>={response.json()["info"]["version"]}')


def _get_requirement(s: str) -> Requirement:  # pragma: no cover
    return Requirement(s)


def _requirement_has_version(r: Requirement) -> bool:  # pragma: no cover
    return r.specifier is not None and r.specifier != ""


def _get_full_requirement_spec(package: str) -> Requirement:  # pragma: no cover
    r = _get_requirement(package)
    if not _requirement_has_version(r):
        r.specifier = _get_package_version_from_pypi(r)
    return r


def _get_reqs_from_pyproject_toml(
    pyproject_toml: Dict,
    dependency_group_name: Optional[str] = None,
) -> Dict[str, Requirement]:
    project = pyproject_toml["project"]
    dep_list = project["dependencies"]
    if dependency_group_name:
        dep_list = project["optional-dependencies"][dependency_group_name]
    return {Requirement(d).name: Requirement(d) for d in dep_list}


def _get_requested_reqs(
    packages: List[str],
) -> Dict[str, Requirement]:  # pragma: no cover
    result = dict()
    for package in packages:
        r = _get_full_requirement_spec(package=package)
        result[r.name] = r
    return result


class TomlEncoder(toml.TomlEncoder):  # pragma: no cover
    def __init__(self, _dict: type = dict, preserve: bool = False) -> None:
        super().__init__(_dict=_dict, preserve=preserve)
        self.dump_funcs[str] = self._dump_str

    def dump_list(self, v: Iterable[Any]) -> str:
        end = "\n"
        leading_space = " " * 4
        retval = f"[{end}"
        for u in v:
            retval += leading_space + str(self.dump_value(u)) + f",{end}"
        retval += "]"
        return retval

    def _dump_str(self, v: str) -> str:
        if v.startswith("/("):
            result = "'''\n" + v + "'''"
        else:
            result = toml.encoder._dump_str(v)  # type: ignore
        if result.startswith('"^') or result.endswith('$"'):
            result = "'" + result + "'"
            result = result.replace('"', "")
        return result.replace("\\\\", "\\")


def _update_pyproject_toml(
    packages: List[str],
    dependency_action: DepencencyAction,
    dependency_group_name: Optional[str] = None,
) -> None:  # pragma: no cover
    pyproject_toml = toml.load(PYPROJECT_TOML_FILENAME)

    current_deps = _get_reqs_from_pyproject_toml(
        pyproject_toml=pyproject_toml,
        dependency_group_name=dependency_group_name,
    )
    requested_deps = _get_requested_reqs(
        packages=packages,
    )

    result = dict()
    if dependency_action == DepencencyAction.add:
        # for all the add requests, check if the dependency is already installed
        # if the dependency is already installed, remove it from the current deps
        # add the requested dependency to the result
        # add the remaining current deps to the result
        # if there are no requested deps, then just re-install the current deps
        for name, requested_dep in requested_deps.items():
            if name in current_deps:
                current_deps.pop(name)
            result[name] = requested_dep
        result.update(current_deps)
    elif dependency_action == DepencencyAction.remove:
        # for all the remove requests, check if the dependency is already installed
        # if the dependency is installed, remove it from the current deps
        for name, current_dep in current_deps.items():
            if name not in requested_deps:
                result[name] = current_dep
    else:
        raise ValueError(
            f"Invalid dependency action: {dependency_action}. "
            f"Valid options are: {DepencencyAction.add}, "
            f"{DepencencyAction.remove}"
        )

    result_list = sorted([str(v) for v in result.values()])

    if dependency_group_name is not None:
        pyproject_toml["project"]["optional-dependencies"][
            dependency_group_name
        ] = result_list
    else:
        pyproject_toml["project"]["dependencies"] = result_list

    with open(PYPROJECT_TOML_FILENAME, "w") as f:
        toml.dump(pyproject_toml, f, encoder=TomlEncoder())


def _snok_sources() -> List[str]:  # pragma: no cover
    """Returns the list of sources to lint with mypy, black,
    and ruff from pyproject.toml."""
    pyproject_toml = toml.load(PYPROJECT_TOML_FILENAME)
    return pyproject_toml["tool"]["snok"]["sources"]


def _get_project_name() -> str:  # pragma: no cover
    pyproject_toml = toml.load(PYPROJECT_TOML_FILENAME)
    return pyproject_toml["project"]["name"]


def _bump_version_string(
    version: str, bump: Optional[BumpType] = None
) -> str:  # pragma: no cover
    """Bump a version string.

    Args:
        version (str): The version string to bump.
        bump (str): The type of bump to perform.

    Returns:
        str: The bumped version string.
    """
    from packaging.version import Version

    v = Version(version)
    if bump == BumpType.MAJOR:
        v = Version(f"{v.major + 1}.0.0")
    elif bump == BumpType.MINOR:
        v = Version(f"{v.major}.{v.minor + 1}.0")
    else:
        v = Version(f"{v.major}.{v.minor}.{v.micro + 1}")
    return str(v)


def _install_node_module_requirements() -> None:  # pragma: no cover
    _run_cmd(
        [
            "npm",
            "install",
            "-D",
            "tailwindcss",
            "@tailwindcss/forms",
            "prettier",
        ]
    )



import time

from fastapi import APIRouter

from {{ __template_name }}._modal import Cache, Queue, Worker
from {{ __template_name }}.kit.routers import _APIRoute
from {{ __template_name }}.logger import log

router = APIRouter(
    route_class=_APIRoute,
    tags=["modal_testing"],
)

#
#   NOTE: This module can be safely deleted. It's only purpose is to demo
#   modal functionality. You can repurpose these functions for your own apps.
#


@router.get("/cache", response_model=None)
async def _test_cache() -> None:
    get = await Cache.get("test")
    log.info(f"get: {get}")
    contains = await Cache.contains("test")
    log.info(f"contains: {contains}")
    await Cache.put("test", "test")
    await Cache.put("test", "toast")
    get = await Cache.get("test")
    log.info(f"get: {get}")
    contains = await Cache.contains("test")
    log.info(f"contains: {contains}")
    await Cache.pop("test")
    get = await Cache.get("test")
    log.info(f"get: {get}")
    contains = await Cache.contains("test")
    log.info(f"contains: {contains}")
    log.info("popping again...")
    await Cache.pop("test")


@router.get("/queue", response_model=None)
async def _test_queue() -> None:
    get = await Queue.get()  # type: ignore
    log.info(f"get: {get}")
    await Queue.put("test")  # type: ignore
    get = await Queue.get()  # type: ignore
    log.info(f"get: {get}")
    await Queue.put_many(["test", "toast"])  # type: ignore
    get = await Queue.get_many(2)  # type: ignore
    log.info(f"get: {get}")
    await Queue.put_many(["test", "toast"])  # type: ignore
    get = await Queue.get_many(3)  # type: ignore
    log.info(f"get: {get}")


@router.get("/worker", response_model=None)
async def _test_worker(backgrounded: bool = False) -> None:
    async def thing_doer() -> None:
        log.info("doing things")
        time.sleep(30)
        log.info("done doing things")
        return

    log.info("running things")
    await Worker.run(
        stub_name="{{ __template_name }}",
        func_name="_run",
        backgrounded=backgrounded,
        func=thing_doer,
    )
    log.info("done running things")

    # NOTE: this is a test of the worker running a function that is not
    #       defined in the stub, when a function is not found in the stub
    #       modal will raise an exception
    # log.info("running things")
    # await Worker.run(
    #     stub_name="{{ __template_name }}",
    #     func_name="_run_worker",
    #     backgrounded=backgrounded,
    #     func=thing_doer,
    # )
    # log.info("done running things")



import os
import sys

from {{ __template_name }}.const import ENV
from pydantic import BaseSettings, Field, PostgresDsn


class Settings(BaseSettings):
    #
    #   Required
    #
    ENV: str = Field(
        ...,
        env="ENV",
        description=f"Environment, can be one of {list(map(lambda x: x.value, ENV))}",
    )
    API_SECRET_KEY: str = Field(
        ...,
        env="API_SECRET_KEY",
        description="The API secret key.",
    )
    APP_URL: str = Field(
        ...,
        env="APP_URL",
        description="The root url.",
    )
    #
    #   Defaults
    #
    LOG_LEVEL: str = Field(
        "WARNING",
        env="LOG_LEVEL",
        description="Logging level, can be one of DEBUG, "
        "INFO, WARNING, ERROR, CRITICAL",
    )
    #
    #   Postgres
    #
    POSTGRES_SCHEME: str = Field(
        "postgresql+asyncpg",
        env="POSTGRES_SCHEME",
        description="The postgres scheme.",
    )
    POSTGRES_USER: str = Field(
        "{{ __template_name }}",
        env="POSTGRES_USER",
        description="The postgres user.",
    )
    POSTGRES_PASSWORD: str = Field(
        "{{ __template_name }}",
        env="POSTGRES_PASSWORD",
        description="The postgres password.",
    )
    POSTGRES_HOST: str = Field(
        "127.0.0.1",
        env="POSTGRES_HOST",
        description="The postgres host.",
    )
    POSTGRES_PORT: int = Field(
        5432,
        env="POSTGRES_PORT",
        description="The postgres port.",
    )
    POSTGRES_DATABASE: str = Field(
        "{{ __template_name }}_development",
        env="POSTGRES_DATABASE",
        description="The postgres database name.",
    )
    DB_POOL_SIZE: int = Field(
        0,
        env="DB_POOL_SIZE",
        description="The database pool size. 0 indicates no limit.",
    )
    DB_MAX_OVERFLOW: int = Field(
        -1,
        env="DB_MAX_OVERFLOW",
        description="The database max overflow. -1 indicates no limit.",
    )
    DB_POOL_PRE_PING: bool = Field(
        True,
        env="DB_POOL_PRE_PING",
        description="Check database connection before using it.",
    )

    @property
    def db_url(self) -> str:
        return PostgresDsn.build(
            scheme=self.POSTGRES_SCHEME,
            user=self.POSTGRES_USER,
            password=self.POSTGRES_PASSWORD,
            host=self.POSTGRES_HOST,
            port=str(self.POSTGRES_PORT),
            path=f"/{self.POSTGRES_DATABASE}",
        )

    class Config:
        env_file = ".env"
        env_file_encoding = "utf-8"



def _set_settings() -> Settings:
    _env = ENV(os.environ["ENV"].lower()) if os.getenv("ENV") else ENV.dev
    # override for test
    if os.getenv("_", "").endswith("pytest") or "pytest" in "".join(sys.argv):
        _env = ENV.test
    Settings.Config.env_file = f".env.{_env.value}"
    return Settings()  # type: ignore


settings = _set_settings()



import logging

import structlog
from {{ __template_name }}.config import settings


def _configure_root_logger() -> None:
    root_logger = logging.getLogger()
    handler = logging.StreamHandler()
    root_logger.setLevel(settings.LOG_LEVEL.upper())
    root_logger.addHandler(handler)


def _configure_global_structlog_defaults() -> None:
    structlog.configure(
        processors=[
            structlog.stdlib.filter_by_level,
            structlog.contextvars.merge_contextvars,
            structlog.processors.add_log_level,
            structlog.dev.set_exc_info,
            structlog.processors.TimeStamper(fmt="iso"),
            structlog.processors.JSONRenderer(),
        ],
        wrapper_class=structlog.make_filtering_bound_logger(logging.NOTSET),
        context_class=dict,
        logger_factory=structlog.stdlib.LoggerFactory(),
        cache_logger_on_first_use=False,
    )


def _setup_logging() -> None:
    _configure_root_logger()
    _configure_global_structlog_defaults()


_setup_logging()
log = structlog.get_logger()



import json
from typing import Callable
from uuid import uuid4

from {{ __template_name }}._types import RequestLoggerMessage, ResponseLoggerMessage, Scope
from {{ __template_name }}.logger import log
from fastapi import Request, Response
from fastapi.routing import APIRoute


class _APIRoute(APIRoute):
    """_APIRoute.

    _APIRoute is a custom APIRoute class that adds a background task to the
    response to log request and response data.
    """

    def get_route_handler(self) -> Callable:
        original_route_handler = super().get_route_handler()

        async def _log(
            req: RequestLoggerMessage,
            res: ResponseLoggerMessage,
        ) -> None:
            rid = {"rid": str(uuid4()).replace("-", "")}
            _res = {"res": json.loads(res.json())}
            _req = {"req": json.loads(req.json())}
            log.debug({**_res, **_req, **rid})

        async def custom_route_handler(request: Request) -> Response:
            body = await request.body() or None
            form = await request.form()
            params = request.query_params._dict
            req = RequestLoggerMessage(
                body=body,
                form=dict(form),
                params=params,
                scope=Scope(**dict(request.scope)),
            )
            response = await original_route_handler(request)
            res = ResponseLoggerMessage(
                status_code=response.status_code,
                raw_headers=response.raw_headers,
                body=response.body,
            )
            await _log(req=req, res=res)
            return response

        return custom_route_handler



from {{ __template_name }}.kit.db import RecordModel

class {{ __template_plural_namespace_caps }}(RecordModel, table=True):
    __tablename__ = "{{ __template_plural_namespace }}"
    {% for field in __template_fields %}
    {{ field[0] }}: {{ field[1] }}{% endfor %}



from typing import Any, Callable, List, Optional

from {{ __template_name }}.config import Settings
from {{ __template_name }}.logger import log
from fastapi import FastAPI
from modal import Dict, Function, Image, Secret, Stub, asgi_app
from modal import Queue as ModalQueue

stub = Stub(name="{{ __template_name }}")
Settings.Config.env_file = ".env.prod"
stub["env"] = Secret.from_dict(
    {str(k): str(v) for k, v in Settings().dict().items()}  # type: ignore
)

_kv = Dict.new()
stub._kv = _kv
stub._kv.persisted(label="{{ __template_name }}_kv")

_queue = ModalQueue.new()
stub._queue = _queue
stub._queue.persisted(label="{{ __template_name }}_queue")

image = Image.debian_slim().pip_install_from_pyproject("pyproject.toml")


@stub.function(
    image=image,
    secret=stub["env"],
)
@asgi_app(
    label="{{ __template_name }}",
)
def _app() -> FastAPI:
    from {{ __template_name }}.app import app

    return app


class Cache:
    @staticmethod
    async def get(key: str) -> Any:
        try:
            return await stub.app._kv.get.aio(key)
        except KeyError as e:
            log.error(f"Cache.get: key ({key}) not found")
            return None

    @staticmethod
    async def contains(key: str) -> Any:
        return await stub.app._kv.contains.aio(key)

    @staticmethod
    async def put(key: str, value: Any) -> None:
        return await stub.app._kv.put.aio(key, value)

    @staticmethod
    async def pop(key: str) -> None:
        try:
            return await stub.app._kv.pop.aio(key)
        except KeyError as e:
            log.error(f"Cache.pop: key ({key}) not found")
            return None


class Queue:
    @staticmethod
    async def get(block: bool = False, timeout: Optional[float] = None) -> Any:
        return await stub.app._queue.get.aio(block=block, timeout=timeout)

    @staticmethod
    async def get_many(
        n: int, block: bool = False, timeout: Optional[float] = None
    ) -> List[Any]:
        return await stub.app._queue.get_many.aio(
            n_values=n, block=block, timeout=timeout
        )

    @staticmethod
    async def put(value: Any) -> None:
        return await stub.app._queue.put.aio(v=value)

    @staticmethod
    async def put_many(values: List[Any]) -> None:
        return await stub.app._queue.put_many.aio(vs=values)


class Worker:
    @staticmethod
    async def run(
        stub_name: str,
        func_name: str,
        backgrounded: bool = True,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        worker = await Function.lookup.aio(stub_name, func_name)
        if backgrounded:
            await worker.spawn.aio(*args, **kwargs)
        else:
            await worker.call.aio(*args, **kwargs)


@stub.function(
    image=image,
    secret=stub["env"],
)
async def _run(func: Callable, *args: Any, **kwargs: Any) -> None:
    await func(*args, **kwargs)



from fastapi import APIRouter

from {{ __template_name }}.const import API_V0
from {{ __template_name }}.kit.routers import _APIRoute
from {{ __template_name }}.modal_testing.router import router as modal_testing_router

router = APIRouter(
    prefix=API_V0,
    route_class=_APIRoute,
)

router.include_router(modal_testing_router)



import asyncio
from typing import AsyncGenerator, Generator
from urllib.parse import urlparse

import asyncpg
import pytest
import pytest_asyncio
from asgi_lifespan import LifespanManager
from fastapi import Request
from httpx import AsyncClient
from sqlalchemy.orm import sessionmaker
from sqlmodel.ext.asyncio.session import AsyncSession

from {{ __template_name }}.kit.db import RecordModel
from {{ __template_name }}.database import async_db_engine
from {{ __template_name }}.logger import log
from {{ __template_name }}.app import app
from {{ __template_name }}.config import settings

TEST_BASE_URL = "http://testserver:8001"


@pytest_asyncio.fixture(scope="session")
async def client() -> AsyncGenerator:
    async with AsyncClient(
        app=app,
        base_url=TEST_BASE_URL,
    ) as client, LifespanManager(app):
        yield client


@pytest.fixture(scope="session", autouse=True)
def event_loop(request: Request) -> Generator:
    loop = asyncio.get_event_loop_policy().new_event_loop()
    yield loop
    loop.close()


@pytest_asyncio.fixture(scope="function", autouse=True)
async def async_db_session() -> AsyncGenerator:
    session = sessionmaker(
        async_db_engine, class_=AsyncSession, expire_on_commit=False
    )

    async with session() as s:
        async with async_db_engine.begin() as conn:
            await conn.run_sync(RecordModel.metadata.create_all)
        yield s

    async with async_db_engine.begin() as conn:
        await conn.run_sync(RecordModel.metadata.drop_all)

    await async_db_engine.dispose()


@pytest_asyncio.fixture(scope="session", autouse=True)
async def setup_database() -> AsyncGenerator:
    parsed_url = urlparse(settings.db_url)
    database = parsed_url.path[1:]
    host = parsed_url.hostname
    port = parsed_url.port
    user = parsed_url.username
    password = parsed_url.password

    try:
        conn = await asyncpg.connect(user=user, password=password, host=host, port=port)
        await conn.execute(f"CREATE DATABASE {database};")
    except asyncpg.exceptions.DuplicateDatabaseError:  # pragma: no cover
        log.debug("Database already exists, skipping creation.")
    finally:
        await conn.close()

    yield

    try:
        conn = await asyncpg.connect(user=user, password=password, host=host, port=port)
        await conn.execute(f"DROP DATABASE {database};")
    except asyncpg.exceptions.DuplicateDatabaseError:  # pragma: no cover
        log.debug("Database already exists, skipping creation.")
    finally:
        await conn.close()
